@@ -17,6 +17,7 @@ from playwright.async_api import (
     Response as PlaywrightResponse,
     Route,
 )
+from playwright_stealth import stealth_async
 from scrapy import Spider, signals
 from scrapy.crawler import Crawler
 from scrapy.http import Request, Response
@@ -61,6 +62,16 @@ def _make_response_logger(context_name: str) -> Callable:
     return _log_request
 
 
+def _get_page_method_result(page: Page, pm: PageMethod):
+    try:
+        method = getattr(page, pm.method)
+    except AttributeError:
+        logger.warning(f"Ignoring {repr(pm)}: could not find method")
+        return None
+    else:
+        return method(*pm.args, **pm.kwargs)
+
+
 class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     def __init__(self, crawler: Crawler) -> None:
         super().__init__(settings=crawler.settings, crawler=crawler)
@@ -163,6 +174,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         context = self.contexts.get("default")
 
         page = await context.new_page()
+        await stealth_async(page)
         self.stats.inc_value("playwright/page_count")
         logger.debug(
             "[Context=%s] New page created, page count is %i (%i for all contexts)",
@@ -173,8 +185,6 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         if self.default_navigation_timeout is not None:
             page.set_default_navigation_timeout(self.default_navigation_timeout)
 
-        # page.on("close", self._make_close_page_callback("default"))
-        # page.on("crash", self._make_close_page_callback("default"))
         page.on("request", _make_request_logger("default"))
         page.on("response", _make_response_logger("default"))
         page.on("request", self._increment_request_stats)
@@ -245,14 +255,15 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             logger.exception(f"[{request.url}] Error while downloading request")
             return Response('', status=504)
         else:
+            return result
+        finally:
             try:
                 if page is not None:
                     if not page.is_closed():
                         await page.close()
             except:
                 logger.error(f"[{request.url}] Error on page close")
-            return result
-        finally:
+                raise
             if self.context_semaphores['default'].locked():
                 self.context_semaphores['default'].release()
 
@@ -294,25 +305,30 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     async def _apply_page_methods(self, page: Page, request: Request) -> None:
         page_methods = request.meta.get("playwright_page_methods") or ()
 
-        if not page_methods and "playwright_page_coroutines" in request.meta:
-            page_methods = request.meta["playwright_page_coroutines"]
-            warnings.warn(
-                "The 'playwright_page_coroutines' request meta key is deprecated,"
-                " please use 'playwright_page_methods' instead.",
-                category=ScrapyDeprecationWarning,
-                stacklevel=1,
-            )
-
         if isinstance(page_methods, dict):
             page_methods = page_methods.values()
         for pm in page_methods:
-            if isinstance(pm, PageMethod):
-                try:
-                    method = getattr(page, pm.method)
-                except AttributeError:
-                    logger.warning(f"Ignoring {repr(pm)}: could not find method")
-                else:
-                    result = method(*pm.args, **pm.kwargs)
+            if isinstance(pm, dict):
+                pm_group = pm.popitem()
+                pm_group_name = pm_group[0]
+                pm_group_items = pm_group[1]
+                if pm_group_name == 'wait_for_any':
+                    timeouts = 0
+                    group_length = len(pm_group_items)
+                    for pgi in pm_group_items:
+                        result = _get_page_method_result(page, pgi)
+                        if result is not None:
+                            try:
+                                pgi.result = await result if isawaitable(result) else result
+                                await page.wait_for_load_state(timeout=self.default_navigation_timeout)
+                                return
+                            except TimeoutError:
+                                timeouts += 1
+                                if timeouts == group_length:
+                                    raise
+            elif isinstance(pm, PageMethod):
+                result = _get_page_method_result(page, pm)
+                if result is not None:
                     pm.result = await result if isawaitable(result) else result
                     await page.wait_for_load_state(timeout=self.default_navigation_timeout)
             else:
